@@ -1,30 +1,44 @@
 # Import stuff
 import logging
-from typing import Callable, Union, Optional, Sequence, Dict, List, Tuple
-import torch
+from contextlib import contextmanager
+from dataclasses import dataclass
+from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union
+
 import torch.nn as nn
 import torch.utils.hooks as hooks
 
-from functools import *
-
-from dataclasses import dataclass
 
 @dataclass
 class LensHandle:
+    """
+    A dataclass that holds information about a PyTorch hook.
+
+    Attributes:
+        hook (hooks.RemovableHandle): Reference to the hook's RemovableHandle.
+        is_permanent (bool, optional): Indicates if the hook is permanent. Defaults to False.
+        context_level (Optional[int], optional): Context level associated with the hooks context
+            manager for the given hook. Defaults to None.
+    """
+
     hook: hooks.RemovableHandle
     is_permanent: bool = False
-    
+    context_level: Optional[int] = None
+
 
 # %%
 # Define type aliases
 NamesFilter = Optional[Union[Callable[[str], bool], Sequence[str]]]
 
+
 # %%
-# A helper class to get access to intermediate activations (inspired by Garcon)
-# It's a dummy module that is the identity function by default
-# I can wrap any intermediate activation in a HookPoint and get a convenient
-# way to add PyTorch hooks
 class HookPoint(nn.Module):
+    """
+    A helper class to access intermediate activations in a PyTorch model (inspired by Garcon).
+
+    HookPoint is a dummy module that acts as an identity function by default. By wrapping any
+    intermediate activation in a HookPoint, it provides a convenient way to add PyTorch hooks.
+    """
+
     def __init__(self):
         super().__init__()
         self.fwd_hooks: List[LensHandle] = []
@@ -38,34 +52,66 @@ class HookPoint(nn.Module):
     def add_perma_hook(self, hook, dir="fwd") -> None:
         self.add_hook(hook, dir=dir, is_permanent=True)
 
-    def add_hook(self, hook, dir="fwd", is_permanent=False) -> None:
-        # Hook format is fn(activation, hook_name)
-        # Change it into PyTorch hook format (this includes input and output,
-        # which are the same for a HookPoint)
+    def add_hook(
+        self, hook, dir="fwd", is_permanent=False, level=None, prepend=False
+    ) -> None:
+        """
+        Hook format is fn(activation, hook_name)
+        Change it into PyTorch hook format (this includes input and output,
+        which are the same for a HookPoint)
+        If prepend is True, add this hook before all other hooks
+        """
         if dir == "fwd":
 
             def full_hook(module, module_input, module_output):
                 return hook(module_output, hook=self)
 
+            full_hook.__name__ = (
+                hook.__repr__()
+            )  # annotate the `full_hook` with the string representation of the `hook` function
+
             handle = self.register_forward_hook(full_hook)
-            handle = LensHandle(handle, is_permanent)
-            self.fwd_hooks.append(handle)
+            handle = LensHandle(handle, is_permanent, level)
+
+            if prepend:
+                # we could just pass this as an argument in PyTorch 2.0, but for now we manually do this...
+                self._forward_hooks.move_to_end(handle.hook.id, last=False)
+                self.fwd_hooks.insert(0, handle)
+
+            else:
+                self.fwd_hooks.append(handle)
+
         elif dir == "bwd":
             # For a backwards hook, module_output is a tuple of (grad,) - I don't know why.
+
             def full_hook(module, module_input, module_output):
                 return hook(module_output[0], hook=self)
+
+            full_hook.__name__ = (
+                hook.__repr__()
+            )  # annotate the `full_hook` with the string representation of the `hook` function
+
             handle = self.register_full_backward_hook(full_hook)
-            handle = LensHandle(handle, is_permanent)
-            self.bwd_hooks.append(handle)
+            handle = LensHandle(handle, is_permanent, level)
+
+            if prepend:
+                # we could just pass this as an argument in PyTorch 2.0, but for now we manually do this...
+                self._backward_hooks.move_to_end(handle.hook.id, last=False)
+                self.bwd_hooks.insert(0, handle)
+            else:
+                self.bwd_hooks.append(handle)
         else:
             raise ValueError(f"Invalid direction {dir}")
 
-    def remove_hooks(self, dir="fwd", including_permanent=False) -> None:
-
+    def remove_hooks(self, dir="fwd", including_permanent=False, level=None) -> None:
         def _remove_hooks(handles: List[LensHandle]) -> List[LensHandle]:
             output_handles = []
             for handle in handles:
-                if not handle.is_permanent or including_permanent:
+                if including_permanent:
+                    handle.hook.remove()
+                elif (not handle.is_permanent) and (
+                    level is None or handle.context_level == level
+                ):
                     handle.hook.remove()
                 else:
                     output_handles.append(handle)
@@ -107,15 +153,22 @@ class HookedRootModule(nn.Module):
     def __init__(self, *args):
         super().__init__()
         self.is_caching = False
+        self.context_level = 0
 
     def setup(self):
-        # Setup function - this needs to be run in __init__ AFTER defining all
-        # layers
-        # Add a parameter to each module giving its name
-        # Build a dictionary mapping a module name to the module
+        """
+        Sets up model.
+
+        This function must be called in the model's `__init__` method AFTER defining all layers. It
+        adds a parameter to each module containing its name, and builds a dictionary mapping module
+        names to the module instances. It also initializes a hook dictionary for modules of type
+        "HookPoint".
+        """
         self.mod_dict = {}
         self.hook_dict: Dict[str, HookPoint] = {}
         for name, module in self.named_modules():
+            if name == "":
+                continue
             module.name = name
             self.mod_dict[name] = module
             if "HookPoint" in str(type(module)):
@@ -124,36 +177,145 @@ class HookedRootModule(nn.Module):
     def hook_points(self):
         return self.hook_dict.values()
 
-    def remove_all_hook_fns(self, direction="both", including_permanent=False):
+    def remove_all_hook_fns(
+        self, direction="both", including_permanent=False, level=None
+    ):
         for hp in self.hook_points():
-            hp.remove_hooks(direction, including_permanent=including_permanent)
+            hp.remove_hooks(
+                direction, including_permanent=including_permanent, level=level
+            )
 
     def clear_contexts(self):
         for hp in self.hook_points():
             hp.clear_context()
 
-    def reset_hooks(self, clear_contexts=True, direction="both", including_permanent=False):
+    def reset_hooks(
+        self,
+        clear_contexts=True,
+        direction="both",
+        including_permanent=False,
+        level=None,
+    ):
         if clear_contexts:
             self.clear_contexts()
-        self.remove_all_hook_fns(direction, including_permanent)
+        self.remove_all_hook_fns(direction, including_permanent, level=level)
         self.is_caching = False
 
-    def check_and_add_hook(self, hook_point, hook_point_name, hook, dir="fwd", is_permanent=False) -> None:
-        """Override this function to add checks on which hooks should be added"""
-        hook_point.add_hook(hook, dir=dir, is_permanent=is_permanent)
+    def check_and_add_hook(
+        self,
+        hook_point,
+        hook_point_name,
+        hook,
+        dir="fwd",
+        is_permanent=False,
+        level=None,
+        prepend=False,
+    ) -> None:
+        """Runs checks on the hook, and then adds it to the hook point"""
+        self.check_hooks_to_add(
+            hook_point,
+            hook_point_name,
+            hook,
+            dir=dir,
+            is_permanent=is_permanent,
+            prepend=prepend,
+        )
+        hook_point.add_hook(
+            hook, dir=dir, is_permanent=is_permanent, level=level, prepend=prepend
+        )
 
-    def add_hook(self, name, hook, dir="fwd", is_permanent=False) -> None:
+    def check_hooks_to_add(
+        self, hook_point, hook_point_name, hook, dir="fwd", is_permanent=False
+    ) -> None:
+        """Override this function to add checks on which hooks should be added"""
+        pass
+
+    def add_hook(
+        self, name, hook, dir="fwd", is_permanent=False, level=None, prepend=False
+    ) -> None:
         if type(name) == str:
-            self.check_and_add_hook(self.mod_dict[name], name, hook, dir=dir, is_permanent=is_permanent)
+            self.check_and_add_hook(
+                self.mod_dict[name],
+                name,
+                hook,
+                dir=dir,
+                is_permanent=is_permanent,
+                level=level,
+                prepend=prepend,
+            )
         else:
             # Otherwise, name is a Boolean function on names
             for hook_point_name, hp in self.hook_dict.items():
                 if name(hook_point_name):
-                    self.check_and_add_hook(hp, hook_point_name, hook, dir=dir, is_permanent=is_permanent)
+                    self.check_and_add_hook(
+                        hp,
+                        hook_point_name,
+                        hook,
+                        dir=dir,
+                        is_permanent=is_permanent,
+                        level=level,
+                        prepend=prepend,
+                    )
 
     def add_perma_hook(self, name, hook, dir="fwd") -> None:
         self.add_hook(name, hook, dir=dir, is_permanent=True)
 
+    @contextmanager
+    def hooks(
+        self,
+        fwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
+        bwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
+        reset_hooks_end: bool = True,
+        clear_contexts: bool = False,
+    ):
+        """
+        A context manager for adding temporary hooks to the model.
+
+        Args:
+            fwd_hooks: List[Tuple[name, hook]], where name is either the name of a hook point or a
+                Boolean function on hook names and hook is the function to add to that hook point.
+            bwd_hooks: Same as fwd_hooks, but for the backward pass.
+            reset_hooks_end (bool): If True, removes all hooks added by this context manager when the context manager exits.
+            clear_contexts (bool): If True, clears hook contexts whenever hooks are reset.
+
+        Example:
+        --------
+        .. code-block:: python
+
+            >>> with model.hooks(fwd_hooks=my_hooks):
+            >>>     hooked_loss = model(text, return_type="loss")
+        """
+        try:
+            self.context_level += 1
+
+            for name, hook in fwd_hooks:
+                if type(name) == str:
+                    self.mod_dict[name].add_hook(
+                        hook, dir="fwd", level=self.context_level
+                    )
+                else:
+                    # Otherwise, name is a Boolean function on names
+                    for hook_name, hp in self.hook_dict.items():
+                        if name(hook_name):
+                            hp.add_hook(hook, dir="fwd", level=self.context_level)
+            for name, hook in bwd_hooks:
+                if type(name) == str:
+                    self.mod_dict[name].add_hook(
+                        hook, dir="bwd", level=self.context_level
+                    )
+                else:
+                    # Otherwise, name is a Boolean function on names
+                    for hook_name, hp in self.hook_dict:
+                        if name(hook_name):
+                            hp.add_hook(hook, dir="bwd", level=self.context_level)
+            yield self
+        finally:
+            if reset_hooks_end:
+                self.reset_hooks(
+                    clear_contexts, including_permanent=False, level=self.context_level
+                )
+            self.context_level -= 1
+
     def run_with_hooks(
         self,
         *model_args,
@@ -164,41 +326,35 @@ class HookedRootModule(nn.Module):
         **model_kwargs,
     ):
         """
-        fwd_hooks: A list of (name, hook), where name is either the name of
-        a hook point or a Boolean function on hook names and hook is the
-        function to add to that hook point, or the hook whose names evaluate
-        to True respectively. Ditto bwd_hooks
-        reset_hooks_end (bool): If True, all hooks are removed at the end (ie,
-        including those added in this run)
-        clear_contexts (bool): If True, clears hook contexts whenever hooks are reset
-        Note that if we want to use backward hooks, we need to set
-        reset_hooks_end to be False, so the backward hooks are still there - this function only runs a forward pass.
+        Runs the model with specified forward and backward hooks.
+
+        Args:
+            fwd_hooks (List[Tuple[Union[str, Callable], Callable]]): A list of (name, hook), where name is
+                either the name of a hook point or a boolean function on hook names, and hook is the
+                function to add to that hook point. Hooks with names that evaluate to True are added
+                respectively.
+            bwd_hooks (List[Tuple[Union[str, Callable], Callable]]): Same as fwd_hooks, but for the
+                backward pass.
+            reset_hooks_end (bool): If True, all hooks are removed at the end, including those added
+                during this run. Default is True.
+            clear_contexts (bool): If True, clears hook contexts whenever hooks are reset. Default is
+                False.
+            *model_args: Positional arguments for the model.
+            **model_kwargs: Keyword arguments for the model.
+
+        Note:
+            If you want to use backward hooks, set `reset_hooks_end` to False, so the backward hooks
+            remain active. This function only runs a forward pass.
         """
-        try:
-            for name, hook in fwd_hooks:
-                if type(name) == str:
-                    self.mod_dict[name].add_hook(hook, dir="fwd")
-                else:
-                    # Otherwise, name is a Boolean function on names
-                    for hook_name, hp in self.hook_dict.items():
-                        if name(hook_name):
-                            hp.add_hook(hook, dir="fwd")
-            for name, hook in bwd_hooks:
-                if type(name) == str:
-                    self.mod_dict[name].add_hook(hook, dir="bwd")
-                else:
-                    # Otherwise, name is a Boolean function on names
-                    for hook_name, hp in self.hook_dict:
-                        if name(hook_name):
-                            hp.add_hook(hook, dir="bwd")
-            return self.forward(*model_args, **model_kwargs)
-        finally:
-            if reset_hooks_end:
-                if len(bwd_hooks) > 0:
-                    logging.warning(
-                        "WARNING: Hooks were reset at the end of run_with_hooks while backward hooks were set. This removes the backward hooks before a backward pass can occur"
-                    )
-                self.reset_hooks(clear_contexts, including_permanent=False)
+        if len(bwd_hooks) > 0 and reset_hooks_end:
+            logging.warning(
+                "WARNING: Hooks will be reset at the end of run_with_hooks. This removes the backward hooks before a backward pass can occur."
+            )
+
+        with self.hooks(
+            fwd_hooks, bwd_hooks, reset_hooks_end, clear_contexts
+        ) as hooked_model:
+            return hooked_model.forward(*model_args, **model_kwargs)
 
     def add_caching_hooks(
         self,
@@ -213,24 +369,24 @@ class HookedRootModule(nn.Module):
         Args:
             names_filter (NamesFilter, optional): Which activations to cache. Can be a list of strings (hook names) or a filter function mapping hook names to booleans. Defaults to lambda name: True.
             incl_bwd (bool, optional): Whether to also do backwards hooks. Defaults to False.
-            device (_type_, optional): The device to store on. Defaults to CUDA if available else CPU.
+            device (_type_, optional): The device to store on. Defaults to same device as model.
             remove_batch_dim (bool, optional): Whether to remove the batch dimension (only works for batch_size==1). Defaults to False.
             cache (Optional[dict], optional): The cache to store activations in, a new dict is created by default. Defaults to None.
 
         Returns:
             cache (dict): The cache where activations will be stored.
         """
-        if device is None:
-            device = "cuda" if torch.cuda.is_available() else "cpu"
         if cache is None:
             cache = {}
 
         if names_filter is None:
             names_filter = lambda name: True
         elif type(names_filter) == str:
-            names_filter = lambda name: name == names_filter
+            filter_str = names_filter
+            names_filter = lambda name: name == filter_str
         elif type(names_filter) == list:
-            names_filter = lambda name: name in names_filter
+            filter_list = names_filter
+            names_filter = lambda name: name in filter_list
 
         self.is_caching = True
 
@@ -265,28 +421,104 @@ class HookedRootModule(nn.Module):
         **model_kwargs,
     ):
         """
-        Runs the model and returns model output and a Cache object
+        Runs the model and returns the model output and a Cache object.
+
+        Args:
+            *model_args: Positional arguments for the model.
+            names_filter (NamesFilter, optional): A filter for which activations to cache. Accepts None, str,
+                list of str, or a function that takes a string and returns a bool. Defaults to None, which
+                means cache everything.
+            device (str or torch.Device, optional): The device to cache activations on. Defaults to the
+                model device. WARNING: Setting a different device than the one used by the model leads to
+                significant performance degradation.
+            remove_batch_dim (bool, optional): If True, removes the batch dimension when caching. Only
+                makes sense with batch_size=1 inputs. Defaults to False.
+            incl_bwd (bool, optional): If True, calls backward on the model output and caches gradients
+                as well. Assumes that the model outputs a scalar (e.g., return_type="loss"). Custom loss
+                functions are not supported. Defaults to False.
+            reset_hooks_end (bool, optional): If True, removes all hooks added by this function at the
+                end of the run. Defaults to True.
+            clear_contexts (bool, optional): If True, clears hook contexts whenever hooks are reset.
+                Defaults to False.
+            **model_kwargs: Keyword arguments for the model.
+
+        Returns:
+            tuple: A tuple containing the model output and a Cache object.
 
-        model_args and model_kwargs - all positional arguments and keyword arguments not otherwise captured are input to the model
-        names_filter (None or str or [str] or fn:str->bool): a filter for which activations to cache. Defaults to None, which means cache everything.
-        device (str or torch.Device): The device to cache activations on, defaults to model device. Note that this must be set if the model does not have a model.cfg.device attribute. WARNING: Setting a different device than the one used by the model leads to significant performance degradation.
-        remove_batch_dim (bool): If True, will remove the batch dimension when caching. Only makes sense with batch_size=1 inputs.
-        incl_bwd (bool): If True, will call backward on the model output and also cache gradients. It is assumed that the model outputs a scalar, ie. return_type="loss", for predict the next token loss. Custom loss functions are not supported
-        reset_hooks_end (bool): If True, all hooks are removed at the end (ie, both those added in this run *and* any added before!)
-        clear_contexts (bool): If True, clears hook contexts whenever hooks are reset
         """
-        cache_dict = self.add_caching_hooks(
+        cache_dict, fwd, bwd = self.get_caching_hooks(
             names_filter, incl_bwd, device, remove_batch_dim=remove_batch_dim
         )
-        model_out = self(*model_args, **model_kwargs)
 
-        if incl_bwd:
-            model_out.backward()
+        with self.hooks(
+            fwd_hooks=fwd,
+            bwd_hooks=bwd,
+            reset_hooks_end=reset_hooks_end,
+            clear_contexts=clear_contexts,
+        ):
+            model_out = self(*model_args, **model_kwargs)
+            if incl_bwd:
+                model_out.backward()
 
-        if reset_hooks_end:
-            self.reset_hooks(clear_contexts, including_permanent=False)
         return model_out, cache_dict
 
+    def get_caching_hooks(
+        self,
+        names_filter: NamesFilter = None,
+        incl_bwd: bool = False,
+        device=None,
+        remove_batch_dim: bool = False,
+        cache: Optional[dict] = None,
+    ) -> Tuple[dict, list, list]:
+        """Creates hooks to cache activations. Note: It does not add the hooks to the model.
+
+        Args:
+            names_filter (NamesFilter, optional): Which activations to cache. Can be a list of strings (hook names) or a filter function mapping hook names to booleans. Defaults to lambda name: True.
+            incl_bwd (bool, optional): Whether to also do backwards hooks. Defaults to False.
+            device (_type_, optional): The device to store on. Keeps on the same device as the layer if None.
+            remove_batch_dim (bool, optional): Whether to remove the batch dimension (only works for batch_size==1). Defaults to False.
+            cache (Optional[dict], optional): The cache to store activations in, a new dict is created by default. Defaults to None.
+
+        Returns:
+            cache (dict): The cache where activations will be stored.
+            fwd_hooks (list): The forward hooks.
+            bwd_hooks (list): The backward hooks. Empty if incl_bwd is False.
+        """
+        if cache is None:
+            cache = {}
+
+        if names_filter is None:
+            names_filter = lambda name: True
+        elif type(names_filter) == str:
+            filter_str = names_filter
+            names_filter = lambda name: name == filter_str
+        elif type(names_filter) == list:
+            filter_list = names_filter
+            names_filter = lambda name: name in filter_list
+        self.is_caching = True
+
+        def save_hook(tensor, hook):
+            if remove_batch_dim:
+                cache[hook.name] = tensor.detach().to(device)[0]
+            else:
+                cache[hook.name] = tensor.detach().to(device)
+
+        def save_hook_back(tensor, hook):
+            if remove_batch_dim:
+                cache[hook.name + "_grad"] = tensor.detach().to(device)[0]
+            else:
+                cache[hook.name + "_grad"] = tensor.detach().to(device)
+
+        fwd_hooks = []
+        bwd_hooks = []
+        for name, hp in self.hook_dict.items():
+            if names_filter(name):
+                fwd_hooks.append((name, save_hook))
+                if incl_bwd:
+                    bwd_hooks.append((name, save_hook_back))
+
+        return cache, fwd_hooks, bwd_hooks
+
     def cache_all(self, cache, incl_bwd=False, device=None, remove_batch_dim=False):
         logging.warning(
             "cache_all is deprecated and will eventually be removed, use add_caching_hooks or run_with_cache"
