@@ -1,14 +1,13 @@
-from transformer_lens import HookedTransformer
-from transformer_lens import HookedTransformerConfig
 from dataclasses import dataclass
-from typing import Optional, Callable
-from torch.utils.data import Dataset, DataLoader
+from typing import Optional
+
+import torch
 import torch.optim as optim
 import wandb
-import torch
-import torch.nn as nn
+from torch.utils.data import DataLoader, Dataset
 from tqdm.auto import tqdm
-from einops import rearrange
+
+from transformer_lens import HookedTransformer, utils
 
 
 @dataclass
@@ -75,7 +74,7 @@ def train(
         wandb.init(project=config.wandb_project_name, config=vars(config))
 
     if config.device is None:
-        config.device = "cuda" if torch.cuda.is_available() else "cpu"
+        config.device = utils.get_device()
 
     if config.optimizer_name in ["Adam", "AdamW"]:
         # Weight decay in Adam is implemented badly, so use AdamW instead (see PyTorch AdamW docs)
