@@ -1,27 +1,46 @@
 from __future__ import annotations
+
+import inspect
+import json
+import re
+import shutil
+from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
+
+import einops
 import numpy as np
 import torch
 import torch.nn.functional as F
-from datasets.arrow_dataset import Dataset
-import einops
-from transformers import AutoTokenizer
-from typing import Optional, Union, Tuple, List, Dict, Type
 import transformers
-from huggingface_hub import hf_hub_download
-import re
-from rich import print as rprint
 from datasets.arrow_dataset import Dataset
 from datasets.load import load_dataset
+from huggingface_hub import hf_hub_download
+from jaxtyping import Float, Int
+from rich import print as rprint
+from transformers import AutoTokenizer
 
 from transformer_lens import FactoredMatrix
 
 CACHE_DIR = transformers.TRANSFORMERS_CACHE
-import json
-from jaxtyping import Float, Int
+
+
+def select_compatible_kwargs(
+    kwargs_dict: Dict[str, Any], callable: Callable
+) -> Dict[str, Any]:
+    """Return a dict with the elements kwargs_dict that are parameters of callable"""
+    return {
+        k: v
+        for k, v in kwargs_dict.items()
+        if k in inspect.getfullargspec(callable).args
+    }
 
 
 def download_file_from_hf(
-    repo_name, file_name, subfolder=".", cache_dir=CACHE_DIR, force_is_torch=False
+    repo_name,
+    file_name,
+    subfolder=".",
+    cache_dir=CACHE_DIR,
+    force_is_torch=False,
+    **kwargs,
 ):
     """
     Helper function to download files from the HuggingFace Hub, from subfolder/file_name in repo_name, saving locally to cache_dir and returning the loaded file (if a json or Torch object) and the file path otherwise.
@@ -29,20 +48,38 @@ def download_file_from_hf(
     If it's a Torch file without the ".pth" extension, set force_is_torch=True to load it as a Torch object.
     """
     file_path = hf_hub_download(
-        repo_id=repo_name, filename=file_name, subfolder=subfolder, cache_dir=cache_dir
+        repo_id=repo_name,
+        filename=file_name,
+        subfolder=subfolder,
+        cache_dir=cache_dir,
+        **select_compatible_kwargs(kwargs, hf_hub_download),
     )
 
-    # Load to the CPU device if CUDA is not available
-    map_location = None if torch.cuda.is_available() else torch.device('cpu')
-
     if file_path.endswith(".pth") or force_is_torch:
-        return torch.load(file_path, map_location=map_location)
+        return torch.load(file_path, map_location="cpu")
     elif file_path.endswith(".json"):
         return json.load(open(file_path, "r"))
     else:
         print("File type not supported:", file_path.split(".")[-1])
         return file_path
 
+
+def clear_huggingface_cache():
+    """
+    Deletes the Hugging Face cache directory and all its contents.
+
+    This function deletes the Hugging Face cache directory, which is used to store downloaded models and their associated files. Deleting the cache directory will remove all the downloaded models and their files, so you will need to download them again if you want to use them in your code.
+
+    Parameters:
+    None
+
+    Returns:
+    None
+    """
+    print("Deleting Hugging Face cache directory and all its contents.")
+    shutil.rmtree(CACHE_DIR)
+
+
 def print_gpu_mem(step_name=""):
     print(
         f"{step_name} ~ {np.round(torch.cuda.memory_allocated()/2e30, 2)} GiB allocated on GPU."
@@ -55,8 +92,10 @@ def get_corner(tensor, n=3):
         return tensor[tuple(slice(n) for _ in range(tensor.ndim))]
     elif isinstance(tensor, FactoredMatrix):
         return tensor[tuple(slice(n) for _ in range(tensor.ndim))].AB
+
+
 def to_numpy(tensor):
-    """ 
+    """
     Helper function to convert a tensor to a numpy array. Also works on lists, tuples, and numpy arrays.
     """
     if isinstance(tensor, np.ndarray):
@@ -114,7 +153,9 @@ def lm_accuracy(
         return correct_matches.sum() / correct_matches.numel()
 
 
-def gelu_new(input: Float[torch.Tensor, "batch pos d_mlp"]) -> Float[torch.Tensor, "batch pos d_mlp"]:
+def gelu_new(
+    input: Float[torch.Tensor, "batch pos d_mlp"]
+) -> Float[torch.Tensor, "batch pos d_mlp"]:
     # Implementation of GeLU used by GPT2 - subtly different from PyTorch's
     return (
         0.5
@@ -128,7 +169,9 @@ def gelu_new(input: Float[torch.Tensor, "batch pos d_mlp"]) -> Float[torch.Tenso
     )
 
 
-def gelu_fast(input: Float[torch.Tensor, "batch pos d_mlp"]) -> Float[torch.Tensor, "batch pos d_mlp"]:
+def gelu_fast(
+    input: Float[torch.Tensor, "batch pos d_mlp"]
+) -> Float[torch.Tensor, "batch pos d_mlp"]:
     return (
         0.5
         * input
@@ -136,7 +179,9 @@ def gelu_fast(input: Float[torch.Tensor, "batch pos d_mlp"]) -> Float[torch.Tens
     )
 
 
-def solu(input: Float[torch.Tensor, "batch pos d_mlp"]) -> Float[torch.Tensor, "batch pos d_mlp"]:
+def solu(
+    input: Float[torch.Tensor, "batch pos d_mlp"]
+) -> Float[torch.Tensor, "batch pos d_mlp"]:
     """
     SoLU activation function as described by
     https://transformer-circuits.pub/2022/solu/index.html.
@@ -240,6 +285,7 @@ print(data)
 tokenize_and_concatenate(data, tokenizer, streaming=False, column_name="text")
 """
 
+
 def sample_logits(
     final_logits: Float[torch.Tensor, "batch d_vocab"],
     top_k: Optional[int] = None,
@@ -247,7 +293,7 @@ def sample_logits(
     temperature: float = 1.0,
     freq_penalty: float = 0.0,
     tokens: Optional[Int[torch.Tensor, "batch pos"]] = None,
-) -> Float[torch.Tensor, "batch"]:
+) -> Int[torch.Tensor, "batch"]:
     """
     Sample from the logits, in order to generate text
 
@@ -286,7 +332,7 @@ def sample_logits(
             indices_to_remove = final_logits < top_logits[..., -1].unsqueeze(-1)
             final_logits = final_logits.masked_fill(indices_to_remove, -float("inf"))
         elif top_p is not None:
-            assert 1.0 >= top_p > 0.0, "top_p has to be in [0, 1)"
+            assert 1.0 >= top_p > 0.0, "top_p has to be in (0, 1]"
             sorted_logits, sorted_indices = torch.sort(final_logits, descending=True)
             cumulative_probs = sorted_logits.softmax(dim=-1).cumsum(dim=-1)
             # We round up - we want prob >= top_p not <top_p
@@ -299,13 +345,23 @@ def sample_logits(
                 -1, sorted_indices, sorted_indices_to_remove
             )
             final_logits = final_logits.masked_fill(indices_to_remove, -float("inf"))
+
+        final_logits = final_logits.to(torch.float32)
         return torch.distributions.categorical.Categorical(logits=final_logits).sample()
 
 
 # %%
 # Type alias
 SliceInput: Type = Optional[
-    Union[int, Tuple[int,], Tuple[int, int], Tuple[int, int, int], List[int], torch.Tensor, np.ndarray]
+    Union[
+        int,
+        Tuple[int,],
+        Tuple[int, int],
+        Tuple[int, int, int],
+        List[int],
+        torch.Tensor,
+        np.ndarray,
+    ]
 ]
 """
 An optional type alias for a slice input used in the `ActivationCache` module.
@@ -356,15 +412,15 @@ class Slice:
 
         Args:
             input_slice (SliceInput): The slice to apply. Can be an int, a tuple, a list, a torch.Tensor, or None. If None, do nothing.
-            
+
         Returns:
             Slice: A Slice object that can be applied to a tensor.
-        
+
         Raises:
             ValueError: If the input_slice is not one of the above types.
         """
         if type(input_slice) == tuple:
-            input_slice = slice(*input_slice)
+            input_slice: slice = slice(*input_slice)
             self.slice = input_slice
             self.mode = "slice"
         elif type(input_slice) == int:
@@ -383,10 +439,10 @@ class Slice:
             raise ValueError(f"Invalid input_slice {input_slice}")
 
     def apply(
-        self, 
-        tensor: torch.Tensor, 
+        self,
+        tensor: torch.Tensor,
         dim: int = 0,
-        ) -> torch.Tensor:
+    ) -> torch.Tensor:
         """
         Takes in a tensor and a slice, and applies the slice to the given dimension (supports positive and negative dimension syntax). Returns the sliced tensor.
 
@@ -403,9 +459,9 @@ class Slice:
         return tensor[tuple(slices)]
 
     def indices(
-        self, 
+        self,
         max_ctx: Optional[int] = None,
-        ) -> Union[np.ndarray, np.int64]:
+    ) -> Union[np.ndarray, np.int32, np.int64]:
         """
         Returns the indices when this slice is applied to an axis of size max_ctx. Returns them as a numpy array, for integer slicing it is eg array([4])
 
@@ -419,14 +475,14 @@ class Slice:
             ValueError: If the slice is not an integer and max_ctx is not specified.
         """
         if self.mode == "int":
-            return np.array([self.slice])
+            return np.array([self.slice], dtype=np.int64)
         if max_ctx is None:
             raise ValueError("max_ctx must be specified if slice is not an integer")
-        return np.arange(max_ctx)[self.slice]
+        return np.arange(max_ctx, dtype=np.int64)[self.slice]
 
     def __repr__(
         self,
-        ) -> str:
+    ) -> str:
         return f"Slice: {self.slice} Mode: {self.mode} "
 
 
@@ -439,9 +495,30 @@ def get_act_name(
     layer_type: Optional[str] = None,
 ):
     """
-    Helper function to convert shorthand to an activation name. Pretty hacky, intended to be useful for short feedback loop hacking stuff together, more so than writing good, readable code. But it is deterministic!
+    Helper function to convert shorthand to an activation name. Pretty hacky, intended to be useful for short feedback
+    loop hacking stuff together, more so than writing good, readable code. But it is deterministic!
+
+    Returns a name corresponding to an activation point in a TransformerLens model.
+
+    Args:
+         name (str): Takes in the name of the activation. This can be used to specify any activation name by itself.
+         The code assumes the first sequence of digits passed to it (if any) is the layer number, and anything after
+         that is the layer type.
+
+         Given only a word and number, it leaves layer_type as is.
+         Given only a word, it leaves layer and layer_type as is.
+
+         Examples:
+             get_act_name('embed') = get_act_name('embed', None, None)
+             get_act_name('k6') = get_act_name('k', 6, None)
+             get_act_name('scale4ln1') = get_act_name('scale', 4, 'ln1')
+
+         layer (int, optional): Takes in the layer number. Used for activations that appear in every block.
+
+         layer_type (string, optional): Used to distinguish between activations that appear multiple times in one block.
+
+    Full Examples:
 
-    eg:
     get_act_name('k', 6, 'a')=='blocks.6.attn.hook_k'
     get_act_name('pre', 2)=='blocks.2.mlp.hook_pre'
     get_act_name('embed')=='hook_embed'
@@ -471,23 +548,35 @@ def get_act_name(
     }
 
     act_name_alias = {
-        "attn":"pattern",
-        "attn_logits":"attn_scores",
-        "key":"k",
-        "query":"q",
-        "value":"v",
-        "mlp_pre":"pre",
-        "mlp_mid":"mid",
-        "mlp_post":"post",
+        "attn": "pattern",
+        "attn_logits": "attn_scores",
+        "key": "k",
+        "query": "q",
+        "value": "v",
+        "mlp_pre": "pre",
+        "mlp_mid": "mid",
+        "mlp_post": "post",
     }
 
+    layer_norm_names = ["scale", "normalized"]
+
     if name in act_name_alias:
         name = act_name_alias[name]
 
     full_act_name = ""
     if layer is not None:
         full_act_name += f"blocks.{layer}."
-    if name in ["k", "v", "q", "z", "rot_k", "rot_q", "result", "pattern", "attn_scores"]:
+    if name in [
+        "k",
+        "v",
+        "q",
+        "z",
+        "rot_k",
+        "rot_q",
+        "result",
+        "pattern",
+        "attn_scores",
+    ]:
         layer_type = "attn"
     elif name in ["pre", "post", "mid"]:
         layer_type = "mlp"
@@ -497,10 +586,15 @@ def get_act_name(
     if layer_type:
         full_act_name += f"{layer_type}."
     full_act_name += f"hook_{name}"
+
+    if name in layer_norm_names and layer is None:
+        full_act_name = f"ln_final.{full_act_name}"
     return full_act_name
 
 
-def remove_batch_dim(tensor: Float[torch.Tensor, "1 ..."]) -> Float[torch.Tensor, "..."]:
+def remove_batch_dim(
+    tensor: Float[torch.Tensor, "1 ..."]
+) -> Float[torch.Tensor, "..."]:
     """
     Removes the first dimension of a tensor if it is size 1, otherwise returns the tensor unchanged
     """
@@ -516,7 +610,7 @@ def test_prompt(
     model,
     prepend_space_to_answer: bool = True,
     print_details: bool = True,
-    prepend_bos: bool = True,
+    prepend_bos: Optional[bool] = None,
     top_k: int = 10,
 ):
     """
@@ -571,10 +665,12 @@ def transpose(tensor: Float[torch.Tensor, "... a b"]) -> Float[torch.Tensor, "..
     """
     return tensor.transpose(-1, -2)
 
+
 def composition_scores(
-    left: FactoredMatrix, right: FactoredMatrix, broadcast_dims=True
+    left: "FactoredMatrix", right: "FactoredMatrix", broadcast_dims=True
 ) -> Union[
-    Float[torch.Tensor, "*leading_dims"], Float[torch.Tensor, "*leading_dims_left *T.leading_dims_right"]
+    Float[torch.Tensor, "*leading_dims"],
+    Float[torch.Tensor, "*leading_dims_left_and_right"],
 ]:
     """
     See `HookedTransformer.all_composition_scores` for documentation.
@@ -601,12 +697,12 @@ def composition_scores(
 # %%
 def get_dataset(dataset_name: str, **kwargs) -> Dataset:
     """
-    Returns a small HuggingFace dataset, for easy testing and exploration. Accesses several convenience datasets with 10,000 elements (dealing with the enormous 100GB - 2TB datasets is a lot of effort!). Note that it returns a dataset (ie a dictionary containing all the data), *not* a DataLoader (iterator over the data + some fancy features). But you can easily convert it to a DataLoader. 
-    
+    Returns a small HuggingFace dataset, for easy testing and exploration. Accesses several convenience datasets with 10,000 elements (dealing with the enormous 100GB - 2TB datasets is a lot of effort!). Note that it returns a dataset (ie a dictionary containing all the data), *not* a DataLoader (iterator over the data + some fancy features). But you can easily convert it to a DataLoader.
+
     Each dataset has a 'text' field, which contains the relevant info, some also have several meta data fields
 
     Kwargs will be passed to the huggingface dataset loading function, e.g. "data_dir"
-    
+
     Possible inputs:
     * openwebtext (approx the GPT-2 training data https://huggingface.co/datasets/openwebtext)
     * pile (The Pile, a big mess of tons of diverse data https://pile.eleuther.ai/)
@@ -616,18 +712,100 @@ def get_dataset(dataset_name: str, **kwargs) -> Dataset:
     * wiki (Wikipedia, generated from the 20220301.en split of https://huggingface.co/datasets/wikipedia )
     """
     dataset_aliases = {
-        'openwebtext': 'stas/openwebtext-10k',
-        'owt': 'stas/openwebtext-10k',
-        'pile': 'NeelNanda/pile-10k',
-        'c4': 'NeelNanda/c4-10k',
-        'code': 'NeelNanda/code-10k',
-        'python': 'NeelNanda/code-10k',
-        'c4_code': 'NeelNanda/c4-code-20k',
-        'c4-code': 'NeelNanda/c4-code-20k',
-        'wiki': 'NeelNanda/wiki-10k'
+        "openwebtext": "stas/openwebtext-10k",
+        "owt": "stas/openwebtext-10k",
+        "pile": "NeelNanda/pile-10k",
+        "c4": "NeelNanda/c4-10k",
+        "code": "NeelNanda/code-10k",
+        "python": "NeelNanda/code-10k",
+        "c4_code": "NeelNanda/c4-code-20k",
+        "c4-code": "NeelNanda/c4-code-20k",
+        "wiki": "NeelNanda/wiki-10k",
     }
     if dataset_name in dataset_aliases:
-        dataset = load_dataset(dataset_aliases[dataset_name], split='train', **kwargs)
+        dataset = load_dataset(dataset_aliases[dataset_name], split="train", **kwargs)
     else:
         raise ValueError(f"Dataset {dataset_name} not supported")
     return dataset
+
+
+def is_square(x: torch.Tensor) -> bool:
+    """Checks if `x` is a square matrix."""
+    return x.ndim == 2 and x.shape[0] == x.shape[1]
+
+
+def is_lower_triangular(x: torch.Tensor) -> bool:
+    """Checks if `x` is a lower triangular matrix."""
+    if not is_square(x):
+        return False
+    return x.equal(x.tril())
+
+
+def check_structure(
+    t1: torch.Tensor, t2: torch.Tensor, *, verbose: bool = False
+) -> None:
+    """Validate that the two square tensors have the same structure, i.e.,
+    that the directionality of comparisons points in the same directions both
+    row-wise and column-wise.
+
+    This function is not used anywhere in the code right now, just for debugging tests.
+    """
+    assert t1.ndim == 2
+    assert t1.shape == t2.shape
+    n_rows, n_cols = cast(Tuple[int, int], t1.shape)
+
+    if verbose:
+        print("Checking rows")
+    row_mismatch = []
+    for row_i in range(n_rows - 1):
+        t1_result = t1[row_i].ge(t1[row_i + 1])
+        t2_result = t2[row_i].ge(t2[row_i + 1])
+        if any(t1_result != t2_result):
+            row_mismatch.append(row_i)
+            if verbose:
+                print(f"\trows {row_i}:{row_i + 1}")
+                print(f"\tt1: {t1_result.tolist()}")
+                print(f"\tt2: {t2_result.tolist()}")
+
+    if verbose:
+        print("Checking columns")
+    col_mismatch = []
+    for col_i in range(n_cols - 1):
+        t1_result = t1[:, col_i].ge(t1[:, col_i + 1])
+        t2_result = t2[:, col_i].ge(t2[:, col_i + 1])
+        if any(t1_result != t2_result):
+            col_mismatch.append(col_i)
+            if verbose:
+                print(f"\trows {col_i}:{col_i + 1}")
+                print(f"\tt1: {t1_result.tolist()}")
+                print(f"\tt2: {t2_result.tolist()}")
+    if not row_mismatch and not col_mismatch:
+        print("PASSED")
+    elif row_mismatch:
+        print(f"row mismatch: {row_mismatch}")
+    elif col_mismatch:
+        print(f"column mismatch: {col_mismatch}")
+
+
+def get_device():
+    if torch.cuda.is_available():
+        return torch.device("cuda")
+    if torch.backends.mps.is_available() and torch.backends.mps.is_built():
+        # Parse the PyTorch version to check if it's below version 2.0
+        major_version = int(torch.__version__.split(".")[0])
+        if major_version >= 2:
+            return torch.device("mps")
+
+    return torch.device("cpu")
+
+
+def override_or_use_default_flag(
+    default_flag: bool,
+    override: Optional[bool] = None,
+) -> bool:
+    """
+    Determines which flag to return based on whether an overriding flag is provided.
+    If a not-None overriding flag is provided, it is returned.
+    Otherwise, the global flag is returned.
+    """
+    return override if override is not None else default_flag
